from pandas import DataFrame
from dmf.preprocess.tokenizers import Tokenizer


class Dataset():
    """
    Class representing a dataset.
    """

    def __init__(self):
        """

        Args:
            df: DataFrame
        """

        self.df = DataFrame()
        """Dataframe contain a dataset has been read by Reader"""

    def get_list_features(self):
        """
        Method to get list of features from Dataframe.

        Returns: List
        """
        return self.df.columns.values

    def entities(self):
        """
        Method to get list of entities from Dataframe

        Returns: List
        """

        list_of_entities = []
        for row in self.df[self.get_list_features()].iterrows():
            entity = row[1]
            list_of_entities.append(dict(entity))
        return list_of_entities

    # def initiate_list_of_records(self, records, ids):
    #     for col in self.df.columns[records]:
    #         for id in ids:


class MappingFeature:

    def __init__(self):
        """

        Args:
            join_features: Dict
            feature_left: List
            feature_right: List
        """
        self.join_features = []
        """
        A List stores the feature objects.
        This represents the mapping of features from feature_left and feature_right.
        """

        self.left_features = []
        """A list of left dataset features is used for matching."""

        self.right_features = []
        """A list of right dataset features is used for matching."""


class Feature:

    def __init__(self, field_name, weight,
                 tokenizer = Tokenizer()):
        """

        """
        self.field_name = field_name
        self.weight = weight
        self.tokenizer = tokenizer

    def __hash__(self):
        return hash((self.field_name, self.weight))



import abc
from dmf.structures.data import (
    Dataset,
)


class Writer():
    """
    Abstract class for writing data.
    """

    @abc.abstractmethod
    def __init__(self, dataset=Dataset()):
        """

        Args:
            dataset: Dataset
        """
        self.dataset = dataset

    @abc.abstractmethod
    def write(self):
        """
        Method to write data
        """


class FileWriter(Writer):
    """
    Class writer for write dataset to file.
    """

    def __init__(self, file_name, dataset):
        """

        Args:
            file_name: String
        """
        super(FileWriter, self).__init__(dataset=dataset)
        self.file_name = file_name
        """Name of file to writing the dataset"""


class DBWriter(Writer):
    """
        Class writer for write dataset to database.
    """

    def __init__(self, db_host, db_user, db_passwd, db_name, dataset):
        """

        Args:
            db_host: String
            db_user: String
            db_passwd: String
            db_name: String
        """
        super(DBWriter, self).__init__(dataset=dataset)
        self.db_host = db_host
        """Database host name to connect"""
        self.db_user = db_user
        """Database user name to connect"""
        self.db_passwd = db_passwd
        """Database password to connect"""
        self.db_name = db_name
        """Database name to connect"""

    @abc.abstractmethod
    def connect(self):
        """
        Method for create an engine to connect to the database.
        """
        return

    @abc.abstractmethod
    def write(self, table_name):
        return



import abc


class DataPreprocessor:
    """
    Abstract class to process and standardize data of the dataset.
    """

    def add_matcher(self, matcher):
        """
        Add the match object on the Dataprocess.

        Args:
            matcher: dmf.match.matchers.Matcher

        Returns: None

        """
        self.matcher = matcher
        """The matcher object"""

    @abc.abstractmethod
    def data_preprocess(self):
        """
        Do data preprocessing.
        """
        self.drop_record_with_none_value_in_feature()
        self.id_preprocess()
        self.initiate_match_record()

    def id_preprocess(self):
        """
        Method to process the IDs of two dataset.

        Returns: None

        """
        self.add_id(self.matcher.left_data.df, "left")
        self.add_id(self.matcher.right_data.df, "right")

    def add_id(self, df, side):
        """
        Static method of DataPreprocessor to add new IDs into the dataset if it is not has IDs

        Args:
            dataset: DataFrame
            side: String

        Returns: None

        """
        id_side = "id_" + side
        id_list = ["{}_{}".format(i, side) for i in range(0, len(df))]
        df[id_side] = id_list

    def initiate_match_record(self):
        """
        Initialize the list of records according to the feature is used for matching
        """
        list_records_left = self.matcher.get_records_by_fields(self.matcher.left_data, self.matcher.features_left, 'id_left')
        list_records_right = self.matcher.get_records_by_fields(self.matcher.right_data, self.matcher.features_right, 'id_right')
        self.matcher.records_join = {}
        for i in range(0, len(self.matcher.join_features)):
            join_record = {}
            join_record.update(list_records_left[i])
            join_record.update(list_records_right[i])
            self.matcher.records_join[self.matcher.join_features[i]] = join_record

    def drop_record_with_none_value_in_feature(self):
        """
        remove records with None feature
        """
        self.matcher.left_data.df = self.matcher.left_data.df.dropna(subset=self.matcher.features_left)
        self.matcher.right_data.df = self.matcher.right_data.df.dropna(subset=self.matcher.features_right)



import abc


class Vectorizer:
    """
       Abstract class for convert a collection of record's data to a matrix of features.
    """

    def add_matcher(self, matcher):
        """
                Add the match object on the Vectorizer.

                Args:
                    matcher: dmf.match.matchers.Matcher

                Returns: None

                """
        self.matcher = matcher
        """The matcher object"""

    @abc.abstractmethod
    def vectorize_matrix(self, feature, records):
        """
        Convert list of data to a matrix of features.

        Returns:
        -------
        vectorized_matrix: array
            A matrix after vectorize.
        """

    def vectorize(self):
        """
        Get each data by for loop and call method vectorized_matrix()

        Returns:
        -------
        vectorized_dict: dict
            A dictionary with feature as key and vectorized_matrix as value.
        """
        vectorized_dict = {}
        for feature, records in self.matcher.records_join.items():
            vectorized_dict[feature] = self.vectorize_matrix(feature, records)
        self.matcher.vectorized_dict = vectorized_dict


class FrequencyVectorizer(Vectorizer):

    def __init__(self,
                 max_df=1,
                 min_df=1,
                 stop_words=[],
                 ngram_range=(1, 1)):
        """

        Args:
            max_df: float or int
            min_df: Float or int
            stop_words: list
            ngram_range: array
        """
        self.max_df = max_df
        """ignore terms in vocabulary that have frequency or proportion higher than max_df"""
        self.min_df = min_df
        """ignore terms in vocabulary that have frequency or proportion lower than min_df"""
        self.stop_words = stop_words
        """terms that were ignored"""
        self.ngram_range = ngram_range
        """the lower and upper boundary of n-grams"""
